// Trackball.cpp: Implementierung der Klasse Trackball.
//
//////////////////////////////////////////////////////////////////////

#include "Vector.h"

//#include <windows.h>
#include <GL/gl.h>
#include "Trackball.h"
#include <math.h>
#include <GL/glut.h>

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Trackball::Trackball(int w, int h, float trackballSize)
{
	this->moving=0;
	this->scaling=0;
	this->W = w;
	this->H = h;
	this->scalefactor = 1.0;
	this->curquat[0] = 1.0;
	this->curquat[1] = 0.0;
	this->curquat[2] = 0.0;
	this->curquat[3] = 0.0;
	this->ballsize=trackballSize;
}

Trackball::~Trackball()
{

}


void Trackball::setDimensions(int w, int h)
{
	this->W = w;
	this->H = h;
}


//the only GLUT-specific function
void Trackball::mouse(int button, int state, int x, int y)
{
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		moving = 1;
		beginx = x;
		beginy = y;
		scaling = (glutGetModifiers() & GLUT_ACTIVE_SHIFT);
	}
	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
		moving = 0;
	}
}


//! this function expects that a glPushMatrix() has taken place in the init() function
void Trackball::rotate()
{
	GLfloat m[4][4];

	glPopMatrix();
	glPushMatrix();
	build_rotmatrix(m, curquat);
	glMultMatrixf(&m[0][0]);
	glScalef(scalefactor, scalefactor, scalefactor);
}


void Trackball::motion(int x, int y)
{
	if (scaling) {
		scalefactor = scalefactor * (1.0 + (((float) (beginy - y)) / H));
	}
	else if (moving) {
		float lastquat[4];
		trackball(lastquat,
				  (2.0 * beginx - W) / W,
				  (H - 2.0 * beginy) / H,
				  (2.0 * x - W) / W,
				  (H - 2.0 * y) / H
				 );
		add_quats(lastquat, curquat, curquat);
	}
	else {
		return;
	}

	beginx = x;
	beginy = y;
	//glutPostRedisplay();
}

void Trackball::trackball(float q[4], float p1x, float p1y, float p2x, float p2y)
{
	Vector a;	//axis of rotation
	float t, phi;
	Vector p1, p2;

	if (p1x == p2x && p1y == p2y) {
		/* Zero rotation */
		q[0] = 0.0;
		q[1] = 0.0;
		q[2] = 0.0;
		q[3] = 1.0;
		return;
	}

	p1 = Vector(p1x, p1y, project_to_sphere(TRACKBALLSIZE,p1x,p1y));
	p2 = Vector(p2x, p2y, project_to_sphere(TRACKBALLSIZE,p2x,p2y));
	
	t = (p1-p2).abs() / (2.0*TRACKBALLSIZE);
    // Avoid problems with out-of-control values...
	if (t >  1.0) t =  1.0;
	if (t < -1.0) t = -1.0;

	phi = 2.0 * asin(t);

	//make quaternion
	a = (p2^p1).norm()*sin(phi/2.0);

	q[0] = a._[0];
	q[1] = a._[1];
	q[2] = a._[2];
	q[3] = cos(phi/2.0);
	
	return;
}

float Trackball::project_to_sphere(float r, float x, float y)
{
	float d, t, z;

	d = sqrt(x*x + y*y);
	if (d < r * 0.70710678118654752440) {    /* Inside sphere */
		z = sqrt(r*r - d*d);
	} else {           /* On hyperbola */
		t = r / 1.41421356237309504880;
		z = t*t / d;
	}
	return z;
}


void Trackball::build_rotmatrix(float m[4][4], float q[4])
{
	m[0][0] = 1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]);
	m[0][1] = 2.0 * (q[0] * q[1] - q[2] * q[3]);
	m[0][2] = 2.0 * (q[2] * q[0] + q[1] * q[3]);
	m[0][3] = 0.0;

	m[1][0] = 2.0 * (q[0] * q[1] + q[2] * q[3]);
	m[1][1] = 1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]);
	m[1][2] = 2.0 * (q[1] * q[2] - q[0] * q[3]);
	m[1][3] = 0.0;

	m[2][0] = 2.0 * (q[2] * q[0] - q[1] * q[3]);
	m[2][1] = 2.0 * (q[1] * q[2] + q[0] * q[3]);
	m[2][2] = 1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]);
	m[2][3] = 0.0;

	m[3][0] = 0.0;
	m[3][1] = 0.0;
	m[3][2] = 0.0;
	m[3][3] = 1.0;
}

#define RENORMCOUNT 97

void Trackball::add_quats(float q1[4], float q2[4], float dest[4])
{
	static int count=0;

	Vector vq1 = Vector(q1);
	Vector vq2 = Vector(q2);

	Vector t = vq1*q2[3] + vq2*q1[3] + (vq2^vq1);

	dest[0] = t._[0];
	dest[1] = t._[1];
	dest[2] = t._[2];
	dest[3] = q1[3] * q2[3] - vq1*vq2;

	if (++count > RENORMCOUNT) {
		count = 0;
		// normalize quat
		float mag = sqrt(dest[0]*dest[0] + dest[1]*dest[1] + dest[2]*dest[2] + dest[3]*dest[3]);
		for (int i = 0; i < 4; i++) dest[i] /= mag;
	}
}
