#include "Vector.h"
#include <math.h>

//////////////////////////////////////////////////////////////////////
// Constructors/Destructor
//////////////////////////////////////////////////////////////////////

Vector::Vector()
{
	//_[0]=0;
	//_[1]=0;
	//_[2]=0;
}


Vector::Vector(float x, float y, float z)
{
	_[0]=x;
	_[1]=y;
	_[2]=z;
}


Vector::Vector(float v[3])
{
	_[0]=v[0];
	_[1]=v[1];
	_[2]=v[2];
}


void Vector::set(float x, float y, float z)
{
	_[0]=x;
	_[1]=y;
	_[2]=z;
}


Vector::~Vector()
{
	//nothing to do
}


//////////////////////////////////////////////////////////////////////
// arithmetic operations
//////////////////////////////////////////////////////////////////////

// " x= "-operators

void Vector::operator+=(const Vector &v)
{
	_[0]+=v._[0];
	_[1]+=v._[1];
	_[2]+=v._[2];
}


void Vector::operator*=(double s)
{
	_[0] *= (float) s;
	_[1] *= (float) s;
	_[2] *= (float) s;
}


void Vector::operator/=(double s)
{
	s = 1.0/s;
	_[0] *= (float) s;
	_[1] *= (float) s;
	_[2] *= (float) s;
}


// binary operators

Vector operator+(const Vector &v1, const Vector &v2)
{
	return Vector(v1._[0] + v2._[0], v1._[1] + v2._[1], v1._[2] + v2._[2]);
}


Vector operator-(const Vector &v1, const Vector &v2)
{
	return Vector(v1._[0] - v2._[0], v1._[1] - v2._[1], v1._[2] - v2._[2]);
}


Vector Vector::operator-()
{
	return Vector(-_[0], -_[1], -_[2]);
}


Vector operator*(const Vector &v, double s)
{
	return Vector((float) (v._[0]*s),
		          (float) (v._[1]*s),
		          (float) (v._[2]*s));
}

Vector operator/(const Vector &v, double s)
{
	return v*(1.0/s);
}


double Vector::abs()
{
	return sqrt(_[0]*_[0] + _[1]*_[1] + _[2]*_[2]);
}


//! normalize vector
Vector Vector::norm()
{
	float d = (float) (1/abs());
	return Vector(_[0]*d, _[1]*d, _[2]*d);
}


//! dot product
double operator*(const Vector &v1, const Vector &v2)
{
	return v1._[0]*v2._[0] + v1._[1]*v2._[1] + v1._[2]*v2._[2];
}


//! cross product
Vector operator^(const Vector &a, const Vector &b)
{
	Vector v;
	v._[0] = a._[1] * b._[2] - a._[2]*b._[1];
	v._[1] = a._[2] * b._[0] - a._[0]*b._[2];
	v._[2] = a._[0] * b._[1] - a._[1]*b._[0];
	return v;
}
